import os

from petdb.pcollection import PetCollection

class PetDB:
	"""
	The main class of PetDB.

	The ``PetDB`` class is responsible for storing and managing this database's collections.

	Collection access is provided by forwarding all unknown method calls
	and property access operations to the :py:meth:`PetDB.collection` method
	by implementing :py:meth:`PetDB.__getattr__` and :py:meth:`PetDB.__getitem__`.

	:param root: The root path where the folder for storing will be created
	"""

	__instances = {}

	@classmethod
	def get(cls, root: str = None):
		if root not in cls.__instances:
			PetDB(root)
		return cls.__instances[root]

	def __init__(self, root: str = None):
		if root in self.__class__.__instances:
			raise Exception("This class is a singleton. Please use PetDB.get() instead.")
		self.__class__.__instances[root] = self
		if root is None:
			root = os.getcwd()
		if not os.path.exists(root):
			raise Exception("Root directory does not exist")
		self.__root = os.path.join(root, "petstorage")
		if not os.path.exists(self.__root):
			os.mkdir(self.__root)

	def __getpath(self, name: str):
		return os.path.join(self.__root, f"{name}.json")

	def collection(self, name: str) -> PetCollection:
		"""
		Get access to the specific collection with the given name.

		If the collection hasn't been accessed yet, a new collection instance will be
		created. Otherwise, the previously created collection instance will be returned.

		:param name: The name of the collection.
		:return: :py:class:`PetCollection`
		"""
		return PetCollection.get_instance(self.__getpath(name))

	def collections(self) -> list[str]:
		"""
		Get the names of all collections in the database.

		:returns: a list of collections names
		"""
		return [os.path.basename(path).rsplit(".", 1)[0] for path in PetCollection.instances().keys()]

	def drop_collection(self, name: str):
		"""
		Deletes the collection with the given name

		:param name: The name of the collection to delete
		"""
		if self.__getpath(name) in PetCollection.instances():
			col = PetCollection.get_instance(self.__getpath(name))
			col.clear()
			os.remove(self.__getpath(name))
			PetCollection.instances().pop(self.__getpath(name))

	def drop(self) -> None:
		"""
		Drop all collections from the database. **CANNOT BE REVERSED!**
		"""
		for col in self:
			col.clear()
			os.remove(self.__getpath(col.name))
		PetCollection.instances().clear()

	def reload(self):
		PetCollection.instances().clear()

	def __getattr__(self, name: str) -> PetCollection:
		"""
		Alias for :py:meth:`PetDB.collection`

		:return: :py:class:`PetCollection`
		"""
		return self.collection(name)

	def __getitem__(self, name: str) -> PetCollection:
		"""
		Alias for :py:meth:`PetDB.collection`

		:return: :py:class:`PetCollection`
		"""
		if not isinstance(name, str):
			raise TypeError("Name must be a string")
		return self.collection(name)

	def __len__(self):
		return len(PetCollection.instances())




from petdb.service.pdb import PetServiceDB
from petdb.service.pcollection import PetServiceCollection, PetServiceMutationsChain
from petdb.service.api import DEFAULT_PORT

__all__ = ["PetServiceDB", "PetServiceCollection", "PetServiceMutationsChain", "DEFAULT_PORT"]



from __future__ import annotations

import os
import json
import time
import uuid
from copy import deepcopy
from types import NoneType
from typing import Optional, Iterator, Self, Callable, Iterable, List, Any, Sized

from petdb.pexceptions import QueryException
from petdb.putils import PetUtils, NON_EXISTENT

type i_sort = str | int | Iterable[str | int] | Callable[[Any], Any] | None

class PetArray[T]:

	def __init__(self, data: List[T] = None):
		self._data: list[T] = data[:] if data is not None else []

	def find(self, query: dict | Callable[[T], bool]) -> Optional[T]:
		"""
		Returns the first element that satisfies the provided ``query``. If no values satisfy, ``None`` is returned.

		:param query: A query object that to match against
		:returns: The first matched element or ``None``
		"""
		for doc in self._data:
			if PetUtils.match(doc, query):
				return doc
		return None

	def findall(self, query: dict | Callable[[T], bool]) -> List[T]:
		"""
		Returns all elements that satisfy the provided ``query``. If no values satisfy, the empty list is returned.

		:param query: A query object that selects which documents to include in the result set
		:returns: List of matched documents
		"""
		return [doc for doc in self._data if PetUtils.match(doc, query)]

	def filter(self, query: dict | Callable[[T], bool]) -> PetArray[T]:
		"""Perform filter mutation. Accepts only query object."""
		return PetArray(self.findall(query))

	def contains(self, query: T | dict | Callable[[T], bool]) -> bool:
		"""
		Check whether the collection contains a document matching a query.

		:param query: the query object
		"""
		for doc in self._data:
			if PetUtils.exists(doc, query):
				return True
		return False

	def exists(self, query: T | dict | Callable[[T], bool]) -> bool:
		"""Alias for :py:meth:`PetArray.contains`"""
		return self.contains(query)

	def sort(self, query: i_sort = None, reverse: bool = False) -> Self:
		"""Perform sort mutation. Accepts a path, list of paths and sorting function."""
		if isinstance(query, (str, int, NoneType)):
			query = [query]

		def key(doc):
			res = []
			for field in query:
				value = PetUtils.get(doc, field)
				res.append((value == NON_EXISTENT, value))
			return res

		self._data.sort(key=query if isinstance(query, Callable) else key, reverse=reverse)
		return self

	def map[TP](self, func: Callable[[T], TP]) -> PetArray[TP]:
		"""Perform map mutation. Accepts only callable object."""
		return PetArray([func(doc) for doc in self._data])

	def pick[TP](self, *queries: str | int | Callable[[T], TP]) -> PetArray[TP]:
		"""
		Makes a list of copies of objects consisting of the picked properties.

		:param queries: Queries to pick keys from objects.
		"""
		if len(queries) == 0:
			return PetArray()
		elif len(queries) == 1:
			return PetArray([value for doc in self._data if (value := PetUtils.get(doc, queries[0])) != NON_EXISTENT])
		if any(not isinstance(query, str) for query in queries):
			raise QueryException("Invalid query: pick with many queries supports only string query")
		return PetArray([{
			query.split(".")[-1]: PetUtils.get(doc, query)
			for query in queries
		} for doc in self._data])

	def omit[TP](self, *queries: str) -> PetArray[TP]:
		"""
		Makes a list of copies of objects without the omitted properties.

		:param queries: Queries to omit keys from objects.
		"""
		array = deepcopy(self._data)
		for query in queries:
			*paths, field = query.split(".")
			for doc in array:
				temp = doc
				for path in paths:
					if isinstance(temp, dict) and path in temp:
						temp = temp[path]
					elif isinstance(temp, list) and path.isdigit() and len(temp) > int(path):
						temp = temp[int(path)]
					else:
						break
				else:
					if isinstance(temp, dict) and field in temp:
						del temp[field]
		return PetArray(array)

	def foreach(self, func: Callable[[T], None]) -> None:
		"""
		Go through all elements and call the given function with each.

		:param func: function that takes each element in the array
		"""
		for element in self._data:
			func(element)

	def size(self, query: T | dict | Callable[[T], bool] = None) -> int:
		"""Returns the amount of all documents in the collection"""
		if query is None:
			return len(self._data)
		amount = 0
		for doc in self._data:
			if PetUtils.match(doc, query):
				amount += 1
		return amount

	def length(self, query: T | dict | Callable[[T], bool] = None) -> int:
		"""Returns the amount of all documents in the collection"""
		return self.size(query)

	def insert(self, element: T) -> T:
		"""
		Insert a new element into the array.

		:param element: the element to insert
		:returns: inserted element
		"""
		element = deepcopy(element)
		self._data.append(element)
		return element

	def insert_many(self, elements: Iterable[T]) -> list[T]:
		"""
		Insert multiple elements into the array.

		:param elements: an Iterable of elements to insert
		:returns: a list containing the inserted elements
		"""
		new_elements = deepcopy(list(elements))
		self._data.extend(new_elements)
		return new_elements

	def clear(self) -> List[T]:
		"""
		Removes all elements from the array.

		:return: Removed elements
		"""
		removed = self._data[:]
		self._data.clear()
		self._on_change()
		return removed

	def remove(self, query: dict) -> List[T]:
		"""
		Removes matched documents. Accepts id, query object, list of ids and list of documents.
		Performs clearing if the query is None. Returns removed documents.

		:returns: removed documents
		"""
		removed = PetUtils.remove(self._data, query)
		self._on_change()
		return removed

	def delete(self, query: dict) -> Self:
		"""Calls the remove method and returns self"""
		self.remove(query)
		return self

	def update(self, update: dict, query: dict | Callable[[T], bool] = None) -> Self:
		"""Applies update query to all elements that match the given query"""
		if not isinstance(update, dict):
			raise QueryException("Invalid update query: it should be a dict")
		if query is not None and not isinstance(query, (dict, Callable)):
			raise QueryException("Invalid query: it should be a dict or callable")
		for doc in self._data:
			if query is None or PetUtils.match(doc, query):
				PetUtils.update(doc, deepcopy(update))
		self._on_change()
		return self

	def update_one(self, update: dict, query: dict | Callable[[T], bool] = None) -> Self:
		"""Applies update query to a single element matching the given query"""
		item = self.find(query or {})
		if item is not None:
			PetUtils.update(item, deepcopy(update))
			self._on_change()
		return self

	def unique(self, query: str | int | Callable[[T], Any] = None) -> PetArray[T]:
		"""Removes all repeated elements"""
		keys = []
		result = []
		for element in self._data:
			key = PetUtils.get(element, query)
			if key not in keys:
				keys.append(key)
				result.append(element)
		return PetArray(result)

	def groupby(self, key: str | int | Iterable[str | int] | Callable[[T], Any] = None) -> PetArray[dict]:
		"""
		Groups elements by field accessed by a ``key``.

		:param key: The key or keys to access value or values to group elements by.
		:return: A PetArray of dicts with ``key`` and ``items`` fields.
		:raises QueryException: If the key type is not a string, integer or iterable of strings or integers.
		"""
		if isinstance(key, (str, int, NoneType)):
			get_key = lambda item: PetUtils.get(item, key)
		elif isinstance(key, Iterable):
			get_key = lambda item: tuple(PetUtils.get(item, k) for k in key)
		elif isinstance(key, Callable):
			get_key = lambda item: key(item)
		else:
			raise QueryException("Invalid group by key: it should be a string, integer, iterable, callable or None")
		result = {}
		for item in self._data:
			key_value = get_key(item)
			result.setdefault(key_value, []).append(item)
		return PetArray([{"key": key, "items": items} for key, items in result.items()])

	def join(self, delimiter: str, adapter: Callable[[T], str] = None) -> str:
		"""
		Compose all elements of a PetArray into the single string with the given ``delimiter``.

		:param delimiter: Separator used to join elements.
		:param adapter: Function to convert each element to string.
		:return: The concatenated string
		"""
		if adapter is None:
			adapter = lambda item: str(item)
		return delimiter.join(adapter(item) for item in self._data)

	def reduce[TP](self, reducer: Callable[[TP, T], TP], init: TP = None) -> TP:
		"""
		Executes the given "reducer" callback function on each element of the array,
		passing in the return value from the calculation on the preceding element.
		The final result of running the reducer across all elements of the array is a single value.

		:param reducer: A function to execute for each element in the array
			with the next signature: ``reducer(accumulator: TP, item: T) -> TP``.
			Its return value becomes the value of the ``accumulator`` parameter
			on the next invocation of reducer. For the last invocation,
			the return value becomes the return value of reduce(). The function is called
			with the following arguments:
		:param init: A value to which ``accumulator`` is initialized the first time the callback is called.
			If ``init`` is specified, ``reducer`` starts executing with the first value in the array as ``item``.
			If ``init`` is not specified, ``accumulator`` is initialized to the first value in the array, and
			``reducer`` starts executing with the second value in the array as currentValue. In this case,
			if the array is empty (so that there's no first value to return as ``accumulator``), a ``None`` is returned.
		:return: The value that results from running the ``reducer`` callback to completion over the entire array.
		"""
		result = init if init is not None else self._data[0] if self.size() > 0 else None
		for item in self._data[0 if init is not None else 1:]:
			result = reducer(result, item)
		return result

	def list(self) -> List[T]:
		"""
		Returns all documents stored in the collection as a list

		:returns: a list with all documents.
		"""
		return self._data[:]

	def _on_change(self):
		pass

	def __len__(self):
		return len(self._data)

	def __iter__(self) -> Iterator[T]:
		return iter(self._data)

	def __getitem__(self, item) -> T:
		return self._data[item]

	def __repr__(self):
		return f"<{self.__class__.__name__} size={self.size()}>"

class PetCollection(PetArray[dict]):
	"""
	Represents a single PetDB collection.
	It provides methods for accessing, managing and manipulating documents.

	:param path: Path to the json file where collection's documents are stored.
				It also used to specify the name of the collection.
	"""

	__instances = {}

	@classmethod
	def get_instance(cls, path: str) -> PetCollection:
		if path not in cls.__instances:
			PetCollection(path)
		return cls.__instances[path]["instance"]

	@classmethod
	def instances(cls) -> dict[str, dict]:
		return cls.__instances

	def __init__(self, path: str):
		if path in self.__class__.__instances:
			raise Exception("This class is a singleton. Please use PetCollection.get_instance() instead.")
		self.__class__.__instances[path] = {"instance": self, "created": int(time.time())}
		self.__path = path
		self.name = os.path.basename(path).rsplit(".", 1)[0]
		if os.path.exists(self.__path):
			with open(self.__path, "r", encoding="utf-8") as f:
				super().__init__(json.load(f))
		else: super().__init__([])

	def replace(self, data: list[dict]):
		"""
		Replaces all documents in the collection with the given data.

		:param data: list of documents to replace
		"""
		self._data = data
		self.dump()

	def dump(self) -> None:
		"""Dumps documents into a collection's storage file"""
		with open(self.__path, "w", encoding="utf-8") as f:
			json.dump(self._data, f, indent=4, ensure_ascii=False)

	def save(self) -> None:
		"""Alias for :py:meth:`PetCollection.dump`"""
		self.dump()

	def insert(self, doc: dict) -> dict:
		"""
		Insert a new document into the collection.

		:param doc: the document to insert
		:returns: inserted document
		"""
		if not isinstance(doc, dict):
			raise TypeError("Document must be of type dict")
		if "_id" in doc and self.get(doc["_id"]):
			raise QueryException("Duplicate id")
		doc = deepcopy(doc)
		if "_id" not in doc:
			doc["_id"] = str(uuid.uuid4())
		self._data.append(doc)
		self.dump()
		return doc

	def insert_many(self, docs: Iterable[dict] & Sized) -> list[dict]:
		"""
		Insert multiple documents into the collection.

		:param docs: an Iterable of documents to insert
		:returns: a list containing the inserted documents
		"""
		for doc in docs:
			if not isinstance(doc, dict):
				raise TypeError("Document must be of type dict")
			if "_id" in doc and self.get(doc["_id"]):
				raise QueryException("Duplicate id")
			doc = deepcopy(doc)
			if "_id" not in doc:
				doc["_id"] = str(uuid.uuid4())
			self._data.append(doc)
		self.dump()
		return self._data[-len(docs):]

	def update_one(self, update: dict, query: str | dict | list[str | dict] = None) -> None:
		if isinstance(query, str):
			query = {"_id": query}
		if isinstance(query, list):
			if all(isinstance(item, str) for item in query):
				query = {"_id": {"$in": query}}
			elif all(isinstance(item, dict) for item in query):
				query = {"_id": {"$in": [doc["_id"] for doc in query]}}
			else:
				raise QueryException("Invalid query format: query list should contains only ids or only docs")
		super().update_one(update, query)

	def update(self, update: dict, query: dict | list[str | dict] = None) -> None:
		if isinstance(query, list):
			if all(isinstance(item, str) for item in query):
				query = {"_id": {"$in": query}}
			elif all(isinstance(item, dict) for item in query):
				query = {"_id": {"$in": [doc["_id"] for doc in query]}}
			else:
				raise QueryException("Invalid query format: query list should contains only ids or only docs")
		if query is not None and not isinstance(query, dict):
			raise QueryException("Invalid query type: query should be dict or list")
		super().update(update, query)

	def remove(self, query: str | dict | list[str | dict]) -> list[dict]:
		if isinstance(query, str):
			query = {"_id": query}
		elif isinstance(query, list):
			if all(isinstance(item, str) for item in query):
				query = {"_id": {"$in": query}}
			elif all(isinstance(item, dict) for item in query):
				query = {"_id": {"$in": [doc["_id"] for doc in query]}}
			else:
				raise QueryException("Invalid delete query: it can only be a list of IDs or a list of docs")
		return super().remove(query)

	def get(self, id: str) -> Optional[dict]:
		"""
		Search for the document with the given ``id``.

		:param id: document's id to search
		:returns: the document with the given ``id`` or ``None``
		"""
		return self.find({"_id": id})

	def filter(self, query: dict | Callable[[dict], bool]) -> PetMutable:
		"""Perform filter mutation. Accepts only query object."""
		return PetMutable(self, self.findall(query))

	def sort(self, query: i_sort = None, reverse: bool = False) -> PetMutable:
		return PetMutable(self, self._data).sort(query)

	def _on_change(self):
		self.dump()

	def __repr__(self):
		return f"<{self.__class__.__name__} name={self.name} size={self.size()}>"

class PetMutable(PetArray[dict]):

	def __init__(self, col: PetCollection, data: list[dict]):
		super().__init__(data)
		self.__col = col

	def get(self, id: str) -> Optional[dict]:
		"""
		Search for a document with the given id

		:param id: document's id
		:return: a single document or ``None`` if no matching document is found
		"""
		return self.find({"_id": id})

	def filter(self, query: dict | Callable[[dict], bool]) -> PetMutable:
		"""Perform filter mutation. Accepts only query object."""
		return PetMutable(self.__col, self.findall(query))

	def insert(self, doc: dict) -> dict:
		"""
		Insert a new document into the parent collection.

		:param doc: the document to insert
		:returns: inserted document
		"""
		doc = self.__col.insert(doc)
		self._data.append(doc)
		return doc

	def insert_many(self, docs: Iterable[dict]) -> list[dict]:
		"""
		Insert multiple documents into the collection.

		:param docs: an Iterable of documents to insert
		:returns: a list containing the inserted documents
		"""
		docs = self.__col.insert_many(docs)
		self._data.extend(docs)
		return docs

	def update_one(self, update: dict, query: dict = None) -> None:
		"""Applies update query to a single element matching the given query"""
		super().update_one(update, query)
		self.__col.dump()

	def update(self, update: dict, query: dict = None) -> None:
		"""
		Applies update query to all documents in mutated list
		that matches the given query, affects the original collection
		"""
		super().update(update, query)
		self.__col.dump()

	def clear(self) -> List[dict]:
		"""
		Removes all documents from the current mutable array and removes all of them from the original collection.

		:return: Removed documents
		"""
		removed = self.__col.remove(self._data)
		self._data = []
		return removed

	def remove(self, query: dict) -> List[dict]:
		"""
		Removes matched documents, affects the original collection.
		Accepts id, query object, list of ids and list of documents.
		Performs clearing if the query is None. Returns removed documents.

		:returns: removed documents
		"""
		removed = super().remove(query)
		self.__col.remove(removed)
		return removed

	def __repr__(self):
		return f"<{self.__class__.__name__} name={self.__col.name} size={self.size()}>"




import os
import sys
import hashlib

from petcmd import Commander

from petdb.service.api import DEFAULT_PORT
from petdb.service.server import Server, STORAGE_PATH

commander = Commander()

SERVICE_NAME = "petdb.database.service"
SERVICE_PATH = os.path.join("/etc/systemd/system", SERVICE_NAME)

template = f"""
[Unit]
Description=PetDB Service

[Service]
User={os.environ.get("USER", "root")}
Environment="LD_LIBRARY_PATH=/usr/local/lib"
Environment="PYTHONUNBUFFERED=1"
WorkingDirectory={STORAGE_PATH}
ExecStart={sys.executable} -u -m petdb.service run -p {{port}}
Restart=always

[Install]
WantedBy=multi-user.target
""".strip()

os.makedirs(STORAGE_PATH, exist_ok=True)

def check_current_configuration(service_path: str) -> bool:
	try:
		with open(service_path, "r") as f:
			executable = f.read().split("ExecStart=")[1].split(" ")[0]
		if executable != sys.executable:
			print(f"PetDB service have already been configured with another python executable: {executable}")
			if input(f"Want to replace it with the current one: {sys.executable} (y/n)? ").lower() != "y":
				return False
	except Exception:
		pass
	return True

@commander.command("init", "reinit")
def init_service(port: int = DEFAULT_PORT):
	if os.path.exists(SERVICE_PATH):
		if not check_current_configuration(SERVICE_PATH):
			return
		os.system(f"sudo systemctl stop {SERVICE_NAME}")
		os.system(f"sudo systemctl disable {SERVICE_NAME}")
		os.remove(SERVICE_PATH)
		os.system(f"sudo systemctl daemon-reload")
	with open(SERVICE_PATH, "w") as f:
		f.write(template.format(port=port))
	os.system(f"sudo systemctl daemon-reload")
	os.system(f"sudo systemctl enable {SERVICE_NAME}")
	os.system(f"sudo systemctl start {SERVICE_NAME}")

@commander.command("run")
def run_service(port: int):
	passwords = {}
	for database in os.listdir(STORAGE_PATH):
		path = os.path.join(STORAGE_PATH, database)
		if os.path.isdir(path) and "pw" in os.listdir(path):
			with open(os.path.join(STORAGE_PATH, database, "pw"), "r") as f:
				passwords[database] = f.read().strip()
	Server(port=port, passwords=passwords).run()

@commander.command("create", "modify")
def create_database(name: str, password: str):
	os.makedirs(os.path.join(STORAGE_PATH, name), exist_ok=True)
	with open(os.path.join(STORAGE_PATH, name, "pw"), "w") as f:
		f.write(hashlib.sha256(password.encode("utf-8")).hexdigest())
	os.system(f"sudo systemctl restart {SERVICE_NAME}")

@commander.command("hash")
def hash_password(password: str):
	print(hashlib.sha256(password.encode("utf-8")).hexdigest())

if __name__ == "__main__":
	commander.process()




from petdb.service.api import API, DEFAULT_PORT
from petdb.service.pcollection import PetServiceCollection

class PetServiceDB:

	def __init__(self, name: str, password: str, port: int = DEFAULT_PORT):
		self.__api = API(name, password, port or DEFAULT_PORT)

	def collection(self, name: str) -> PetServiceCollection:
		return PetServiceCollection(name, self.__api)

	def collections(self) -> list[str]:
		return self.__api.collections()

	def drop(self):
		self.__api.drop_db()

	def drop_collection(self, name: str):
		self.__api.drop_collection(name)

	@staticmethod
	def enable_debug_mode(enable: bool = True):
		API.enable_debug_mode(enable)




from petdb.pdb import PetDB
from petdb.pcollection import PetCollection, PetMutable, PetArray
from petdb.putils import NonExistent, NON_EXISTENT
from petdb.pexceptions import QueryException

__version__ = "0.9.0"

__all__ = [
	"PetDB", "PetCollection", "PetMutable", "PetArray",
	"NonExistent", "NON_EXISTENT", "QueryException",
	"__version__"
]




import requests
from petdb import PetDB, PetArray
from petdb import __version__

DEFAULT_PORT = 3944

class API:

	debug_mode: bool = False
	localdb: PetDB | None = None

	def __init__(self, dbname: str, password: str, port: int):
		self.dbname = dbname
		self.password = password
		self.port = port

	def drop_db(self):
		if self.debug_mode:
			return self.localdb.drop()
		return self.__request("/drop")

	def drop_collection(self, name: str):
		if self.debug_mode:
			return self.localdb.drop_collection(name)
		return self.__request(f"/drop/{name}")

	def collections(self) -> list[str]:
		if self.debug_mode:
			return self.localdb.collections()
		return self.__request("/collections")

	def mutate(self, name: str, mutations: list[dict]):
		if self.debug_mode:
			array = self.localdb.collection(name)
			for mutation in mutations:
				array: PetArray = array.__getattribute__(mutation["type"])(*mutation["args"])
			return array.list()
		return self.__request(f"/mutate/{name}", {"mutations": mutations})

	def insert(self, name: str, doc: dict):
		if self.debug_mode:
			return self.localdb.collection(name).insert(doc)
		return self.__request(f"/insert/{name}", {"doc": doc})

	def insert_many(self, name: str, docs: list[dict]):
		if self.debug_mode:
			return self.localdb.collection(name).insert_many(docs)
		return self.__request(f"/insert_many/{name}", {"docs": docs})

	def update_one(self, name: str, update: dict, query: dict):
		if self.debug_mode:
			return self.localdb.collection(name).update_one(update, query)
		return self.__request(f"/update_one/{name}", {"update": update, "query": query})

	def update(self, name: str, update: dict, query: dict):
		if self.debug_mode:
			return self.localdb.collection(name).update(update, query)
		return self.__request(f"/update/{name}", {"update": update, "query": query})

	def remove(self, name: str, query: dict):
		if self.debug_mode:
			return self.localdb.collection(name).remove(query)
		return self.__request(f"/remove/{name}", {"query": query})

	def clear(self, name: str):
		if self.debug_mode:
			return self.localdb.collection(name).clear()
		return self.__request(f"/clear/{name}")

	@classmethod
	def enable_debug_mode(cls, enable: bool = True):
		cls.debug_mode = enable
		cls.localdb = PetDB.get()

	def __request(self, endpoint: str, body: dict = None):
		if self.debug_mode:
			return None
		if body is None:
			body = {}
		body["password"] = self.password
		body["dbname"] = self.dbname
		body["version"] = __version__
		response = requests.post(f"http://127.0.0.1:{self.port}{endpoint}", json=body)
		if response is not None and response.headers.get("content-type") == "application/json":
			return response.json()




import os
import time
import hashlib
import threading
from queue import Queue
from typing import Callable, Any

import uvicorn
from fastapi import FastAPI, Request, Response, status, Body

from petdb import PetDB, PetCollection, PetArray

STORAGE_PATH = "/var/lib/petdb"

app = FastAPI()

@app.post("/collections")
def get_collections(request: Request):
	return request.state.db.collections()

@app.post("/drop")
def drop_collections(request: Request):
	request.state.db.drop()

@app.post("/drop/{name}")
def drop_collection(request: Request, name: str):
	request.state.db.drop_collection(name)

@app.post("/mutate/{name}")
def mutate(request: Request, name: str, mutations: list[dict] = Body(embed=True)):
	array = request.state.db.collection(name)
	for mutation in mutations:
		array: PetArray = array.__getattribute__(mutation["type"])(*mutation["args"])
	return array.list()

@app.post("/insert/{name}")
def insert(request: Request, name: str, doc: dict = Body(embed=True)):
	return request.state.db.collection(name).insert(doc)

@app.post("/insert_many/{name}")
def insert_many(request: Request, name: str, docs: list[dict] = Body(embed=True)):
	return request.state.db.collection(name).insert_many(docs)

@app.post("/update_one/{name}")
def update_one(request: Request, name: str, update: dict = Body(embed=True), query: dict = Body(embed=True)):
	return request.state.db.collection(name).update_one(update, query)

@app.post("/update/{name}")
def update(request: Request, name: str, update: dict = Body(embed=True), query: dict = Body(embed=True)):
	return request.state.db.collection(name).update(update, query)

@app.post("/remove/{name}")
def remove(request: Request, name: str, query: dict = Body(embed=True)):
	return request.state.db.collection(name).remove(query)

@app.post("/clear/{name}")
def clear(request: Request, name: str):
	return request.state.db.collection(name).clear()


class Cache[T]:

	def __init__(self, factory: Callable[..., T]):
		self.factory = factory
		self.instances = {}

	def get(self, key, *args) -> T:
		if key not in self.instances:
			self.instances[key] = self.factory(key, *args)
		return self.instances[key]

class Server:

	LOCK_TIMEOUT = 5 * 60 # 5 min

	def __init__(self, port: int, passwords: dict[str, str]):
		self.passwords = passwords
		self.port = port
		self.db = Cache(self.create_db_object)
		self.tasks = Cache(lambda name: Queue())

	def create_db_object(self, name: str) -> PetDB:
		threading.Thread(target=self.process_requests, args=(name,), daemon=True).start()
		return PetDB.get(os.path.join(STORAGE_PATH, name))

	def run(self):

		@app.middleware("http")
		async def middleware(request: Request, call_next):
			body = await request.json()
			dbname = body.get("dbname")
			password = body.get("password")
			if dbname is None or dbname not in self.passwords:
				return Response(status_code=status.HTTP_400_BAD_REQUEST)
			if password is None or hashlib.sha256(password.encode("utf-8")).hexdigest() != self.passwords[dbname]:
				return Response(status_code=status.HTTP_401_UNAUTHORIZED)
			request.state.db = self.db.get(dbname)
			return self.task(self.tasks.get(dbname), call_next, request)

		# noinspection PyDeprecation
		@app.on_event("startup")
		def start_queue_processor():
			threading.Thread(target=self.cache_monitor, daemon=True).start()

		uvicorn.run(app, host="127.0.0.1", port=self.port)

	def process_requests(self, dbname: str):
		tasks = self.tasks.get(dbname)
		while True:
			task, args, result = tasks.get()
			try:
				result.put(task(*args))
			except Exception:
				tasks.put(None)
			finally:
				tasks.task_done()

	def cache_monitor(self):
		while True:
			print("start cache checking...")
			now = int(time.time())
			instances = PetCollection.instances()
			for path in list(instances.keys()):
				dbname = os.path.relpath(path, STORAGE_PATH).split(os.sep)[0]
				print(f"check {dbname}.{instances[path]["instance"].name}...")
				if now - instances[path]["created"] > 3 * 24 * 3600:
					self.task(self.tasks.get(dbname), self.clear_cache, (instances, path))
			time.sleep(24 * 3600)

	def clear_cache(self, instances: dict[str, dict], path: str):
		print(f"clear {instances[path]["instance"].name}")
		del instances[path]

	def task(self, tasks: Queue, task: Callable, args: Any):
		if not isinstance(args, tuple):
			args = (args,)
		result = Queue(maxsize=1)
		tasks.put((task, args, result))
		return result.get()



import re
from typing import Any, Callable

from petdb.pexceptions import QueryException

class NonExistent:
	"""
	Represents a value of all non-existent fields.
	"""

	def __repr__(self):
		return "[Non-existent-object]"

	def __eq__(self, other):
		return isinstance(other, NonExistent)

NON_EXISTENT = NonExistent()
NON_EXISTENT.__doc__ = "The single instance of :py:class:`NonExistent` class."

class PetUtils:

	OPERATORS = {
		"$eq": lambda q, v: v == q,
		"$ne": lambda q, v: v != q,
		"$lt": lambda q, v: v < q,
		"$<": lambda q, v: v < q,
		"$lte": lambda q, v: v <= q,
		"$<=": lambda q, v: v <= q,
		"$gt": lambda q, v: v > q,
		"$>": lambda q, v: v > q,
		"$gte": lambda q, v: v >= q,
		"$>=": lambda q, v: v >= q,
		"$in": lambda q, v: v in q,
		"$nin": lambda q, v: v not in q,
		"$contains": lambda q, v: q in v,
		"$notcontains": lambda q, v: q not in v,
		"$length": lambda q, v: PetUtils.match(len(v), q),
		"$elemMatch": lambda q, v: any(PetUtils.match(item, q) for item in v),
		"$size": lambda q, v: PetUtils.match(len(v), q),
		"$exists": lambda q, v: q == (v != NON_EXISTENT),
		"$regex": lambda q, v: v and re.search(q, v),
		"$func": lambda q, v: q(v),
		"$where": lambda q, v: q(v),
		"$f": lambda q, v: q(v),
		"$type": lambda q, v: isinstance(v, q),
		"$is": lambda q, v: isinstance(v, q),
		"$and": lambda q, v: all(PetUtils.match(v, query) for query in q),
		"$all": lambda q, v: all(PetUtils.match(v, query) for query in q),
		"$or": lambda q, v: any(PetUtils.match(v, query) for query in q),
		"$any": lambda q, v: any(PetUtils.match(v, query) for query in q),
		"$not": lambda q, v: not PetUtils.match(v, q),
	}

	UPDATE_OPERATORS = {
		"$set": lambda q, doc: [PetUtils.set(doc, key, value) for key, value in q.items()],
		"$unset": lambda q, doc: [PetUtils.unset(doc, key) for key, value in q.items() if value],
		"$inc": lambda q, doc: [PetUtils.inc(doc, key, value) for key, value in q.items()],
		"$push": lambda q, doc: [PetUtils.push(doc, key, value) for key, value in q.items()],
		"$append": lambda q, doc: [PetUtils.push(doc, key, value) for key, value in q.items()],
		"$addToSet": lambda q, doc: [PetUtils.add_to_set(doc, key, value) for key, value in q.items()],
		"$pull": lambda q, doc: [PetUtils.pull(doc, key, query) for key, query in q.items()],
		"$remove": lambda q, doc: [PetUtils.pull(doc, key, query) for key, query in q.items()],
		"$map": lambda q, doc: [PetUtils.set(doc, key, func(PetUtils.get(doc, key))) for key, func in q.items()]
	}

	@classmethod
	def match[T](cls, obj: T, query: dict[str] | Callable[[Any], bool] | T) -> bool:
		if isinstance(query, Callable):
			return query(obj)
		elif not isinstance(query, dict):
			return obj == query

		for key in query:
			if key.startswith("$"):
				if not cls.OPERATORS[key](query[key], obj):
					return False
				continue

			value = cls.get(obj, key)
			if cls.is_operators_query(query[key]):
				for operator in query[key]:
					if not cls.OPERATORS[operator](query[key][operator], value):
						return False
			elif isinstance(value, dict) and isinstance(query[key], dict):
				if not cls.match(value, query[key]):
					return False
			elif value != query[key]:
				return False
		return True

	@classmethod
	def exists[T](cls, obj: T, query: dict[str] | Callable[[T], bool]) -> bool:
		if isinstance(query, Callable):
			return query(obj)
		elif not isinstance(query, dict):
			return obj == query

		for key in query:
			if key.startswith("$"):
				try:
					if not cls.OPERATORS[key](query[key], obj):
						return False
				except TypeError:
					return False
				continue

			value = cls.get(obj, key)
			if isinstance(value, NonExistent):
				return False
			elif cls.is_operators_query(query[key]):
				for operator in query[key]:
					try:
						if not cls.OPERATORS[operator](query[key][operator], value):
							return False
					except TypeError:
						return False
			elif isinstance(value, dict) and isinstance(query[key], dict):
				if not cls.exists(value, query[key]):
					return False
			elif value != query[key]:
				return False
		return True

	@classmethod
	def update(cls, obj: dict, update: dict):
		for operator in update:
			cls.UPDATE_OPERATORS[operator](update[operator], obj)

	@classmethod
	def is_operators_query(cls, query: Any) -> bool:
		if not isinstance(query, dict) or len(query) == 0:
			return False

		is_operators_query = list(query.keys())[0].startswith("$")
		for key in query:
			if is_operators_query != key.startswith("$"):
				raise QueryException("Invalid query: you can't combine operators query with the path objects")
		return is_operators_query

	@classmethod
	def get[T](cls, obj: T, key: str | int | Callable[[T], bool] = None, fill_path: bool = False):
		if key is None:
			if obj is None:
				return NON_EXISTENT
			return obj
		elif isinstance(key, Callable):
			try:
				return key(obj)
			except Exception:
				return NON_EXISTENT
		elif isinstance(key, int):
			if not isinstance(obj, list) or key >= len(obj):
				return NON_EXISTENT
			return obj[key]
		elif key == "&":
			return obj

		for field in key.split("."):
			if isinstance(obj, dict):
				if field not in obj:
					if not fill_path:
						return NON_EXISTENT
					obj[field] = {}
				obj = obj[field]
			elif isinstance(obj, list):
				if not field.isdigit() or int(field) >= len(obj):
					return NON_EXISTENT
				obj = obj[int(field)]
			else:
				return NON_EXISTENT
		return obj

	@classmethod
	def set(cls, obj: dict, key: str | int, value: Any):
		if isinstance(key, int):
			if not isinstance(obj, list):
				raise QueryException("Invalid set query: only lists supports integer keys")
			if key >= len(obj):
				raise QueryException("Invalid set query: index out of range")
			obj[key] = value
			return

		if "." in key:
			obj = cls.get(obj, key.rsplit(".", 1)[0], fill_path=True)
			if not isinstance(obj, (dict, list)):
				raise QueryException(f"Invalid set query: path {key} doesn't exist")
			key = key.rsplit(".", 1)[1]

		if isinstance(obj, dict):
			obj[key] = value
		elif isinstance(obj, list):
			if not key.isdigit():
				raise QueryException("Invalid set query: list index must contains only digits")
			if int(key) >= len(obj):
				raise QueryException("Invalid set query: list index out of range")
			obj[int(key)] = value
		else:
			raise QueryException(f"Invalid set query: path {key} doesn't exist")

	@classmethod
	def unset(cls, obj: dict, key: str) -> Any:
		if "." not in key:
			return obj.pop(key, None)

		obj = cls.get(obj, key.rsplit(".", 1)[0])
		if isinstance(obj, dict):
			return obj.pop(key.rsplit(".", 1)[1], None)

	@classmethod
	def inc(cls, obj: dict | list, key: str, value: Any):
		if "." in key:
			obj = cls.get(obj, key.rsplit(".", 1)[0], fill_path=True)
			if not isinstance(obj, (list, dict)):
				raise QueryException(f"Invalid inc query: path {key} doesn't exist")
			key = key.rsplit(".", 1)[1]

		if isinstance(obj, list):
			if not key.isdigit():
				raise QueryException("Invalid inc query: list index must contains only digits")
			if int(key) >= len(obj):
				raise QueryException("Invalid inc query: list index out of range")
			if not isinstance(obj[int(key)], int):
				raise QueryException("Invalid inc query: you can increment only integer values")
			obj[int(key)] += value
		elif isinstance(obj, dict):
			if key not in obj:
				obj[key] = 0
			if not isinstance(obj[key], int):
				raise QueryException(f"Invalid inc query: you can increment only integer values")
			obj[key] += value
		else:
			raise QueryException(f"Invalid set query: path {key} doesn't exist")

	@classmethod
	def push(cls, obj: dict, key: str | int, value: Any):
		if isinstance(key, int):
			if not isinstance(obj, list):
				raise QueryException("Invalid push query: only lists supports integer keys")
			if key >= len(obj):
				raise QueryException("Invalid push query: index out of range")
			if not isinstance(obj[key], list):
				raise QueryException("Invalid push query: it's impossible to append not to a list")
			obj[key].append(value)
			return

		obj = cls.get(obj, key)
		if obj == NON_EXISTENT:
			raise QueryException(f"Invalid push query: path {key} doesn't exist")
		elif not isinstance(obj, list):
			raise QueryException("Invalid push query: it's impossible to append not to a list")
		obj.append(value)

	@classmethod
	def pull(cls, obj: dict, key: str | int, query: dict):
		if isinstance(key, int):
			if not isinstance(obj, list):
				raise QueryException("Invalid pull query: only lists supports integer keys")
			if key >= len(obj):
				raise QueryException("Invalid pull query: index out of range")
			if not isinstance(obj[key], list):
				raise QueryException("Invalid pull query: it's impossible to append not to a list")
			array = obj[key]
		elif isinstance(key, str):
			array = cls.get(obj, key)
			if array == NON_EXISTENT:
				raise QueryException(f"Invalid pull query: path {key} doesn't exist")
		else:
			raise QueryException(f"Invalid pull query: key should be only an integer or a string")

		if not isinstance(array, list):
			raise QueryException("Invalid pull query: it's impossible to append not to a list")

		for item in array[:]:
			if cls.match(item, query):
				array.remove(item)

	@classmethod
	def add_to_set(cls, obj: dict, key: str | int, value: Any):
		if isinstance(key, int):
			if not isinstance(obj, list):
				raise QueryException("Invalid addToSet query: only lists supports integer keys")
			if key >= len(obj):
				raise QueryException("Invalid addToSet query: index out of range")
			if not isinstance(obj[key], list):
				raise QueryException("Invalid addToSet query: it's impossible to append not to a list")
			if value not in obj[key]:
				obj[key].append(value)
			return

		obj = cls.get(obj, key)
		if obj == NON_EXISTENT:
			raise QueryException(f"Invalid addToSet query: path {key} doesn't exist")
		elif not isinstance(obj, list):
			raise QueryException("Invalid addToSet query: it's impossible to append not to a list")
		if value not in obj:
			obj.append(value)

	@classmethod
	def remove(cls, data: list, query: dict | Callable[[Any], bool]):
		if not isinstance(query, dict):
			raise QueryException("Invalid delete query")
		deleted = []
		for item in data[:]:
			if cls.match(item, query):
				deleted.append(item)
				data.remove(item)
		return deleted




from __future__ import annotations

import uuid
from copy import deepcopy
from typing import Optional, Iterator, Callable, Iterable, Sized

from petdb.pcollection import PetArray
from petdb.pexceptions import QueryException
from petdb.putils import PetUtils
from petdb.service.api import API

class PetServiceMutationsChain:

	def __init__(self, name: str, api: API, mutations: list[dict] = None):
		self.name = name
		self._api = api
		self.__mutations = mutations or []
		self._mutated_data = None

	def filter(self, query: dict) -> PetServiceMutationsChain:
		return self.__mutate("filter", [query])

	def sort(self, query: str | int | Iterable[str | int] = None, reverse: bool = False) -> PetServiceMutationsChain:
		return self.__mutate("sort", [query, reverse])

	def pick(self, *queries: str | int) -> PetServiceMutationsChain:
		return self.__mutate("pick", queries)

	def omit(self, *queries: str | int) -> PetServiceMutationsChain:
		return self.__mutate("omit", queries)

	def unique(self, query: str | int = None) -> PetServiceMutationsChain:
		return self.__mutate("unique", [query])

	def groupby(self, key: str | int | Iterable[str | int] = None) -> PetServiceMutationsChain:
		return self.__mutate("groupby", [key])

	def find(self, query: dict | Callable[[dict], bool]) -> Optional[dict]:
		"""
		Returns the first element that satisfies the provided ``query``. If no values satisfy, ``None`` is returned.

		:param query: A query object that to match against
		:returns: The first matched element or ``None``
		"""
		for doc in self._data:
			if PetUtils.match(doc, query):
				return doc
		return None

	def findall(self, query: dict | Callable[[dict], bool]) -> list[dict]:
		"""
		Returns all elements that satisfy the provided ``query``. If no values satisfy, the empty list is returned.

		:param query: A query object that selects which documents to include in the result set
		:returns: List of matched documents
		"""
		return [doc for doc in self._data if PetUtils.match(doc, query)]

	def join(self, delimiter: str, adapter: Callable[[dict], str] = None) -> str:
		"""
		Compose all elements into the single string with the given ``delimiter``.

		:param delimiter: Separator used to join elements.
		:param adapter: Function to convert each element to string.
		:return: The concatenated string.
		"""
		if adapter is None:
			adapter = lambda item: str(item)
		return delimiter.join(adapter(item) for item in self._data)

	def array(self) -> PetArray[dict]:
		"""
		Returns all documents stored in the collection as a :py:class:`PetArray`

		:returns: :py:class:`PetArray` with containing documents.
		"""
		return PetArray(deepcopy(self._data))

	def map[TP](self, func: Callable[[dict], TP]) -> PetArray[TP]:
		"""Perform map mutation. Accepts only callable object."""
		return PetArray([func(deepcopy(doc)) for doc in self._data])

	def reduce[TP](self, reducer: Callable[[TP, dict], TP], init: TP = None) -> TP:
		"""
		Executes the given "reducer" callback function on each element of the array,
		passing in the return value from the calculation on the preceding element.
		The final result of running the reducer across all elements of the array is a single value.

		:param reducer: A function to execute for each element in the array
			with the next signature: ``reducer(accumulator: TP, item: T) -> TP``.
			Its return value becomes the value of the ``accumulator`` parameter
			on the next invocation of reducer. For the last invocation,
			the return value becomes the return value of reduce(). The function is called
			with the following arguments:
		:param init: A value to which ``accumulator`` is initialized the first time the callback is called.
			If ``init`` is specified, ``reducer`` starts executing with the first value in the array as ``item``.
			If ``init`` is not specified, ``accumulator`` is initialized to the first value in the array, and
			``reducer`` starts executing with the second value in the array as currentValue. In this case,
			if the array is empty (so that there's no first value to return as ``accumulator``), a ``None`` is returned.
		:return: The value that results from running the ``reducer`` callback to completion over the entire array.
		"""
		result = init if init is not None else self._data[0] if self.size() > 0 else None
		for item in self._data[0 if init is not None else 1:]:
			result = reducer(result, item)
		return result

	def foreach(self, func: Callable[[dict], None]) -> None:
		"""
		Go through all elements and call the given function with each.

		:param func: function that takes each element in the array
		"""
		for element in self._data:
			func(element)

	def contains(self, query: dict | Callable[[dict], bool]) -> bool:
		"""
		Check whether the collection contains a document matching a query.

		:param query: the query object
		"""
		for doc in self._data:
			if PetUtils.exists(doc, query):
				return True
		return False

	def exists(self, query: dict | Callable[[dict], bool]) -> bool:
		"""Alias for :py:meth:`PetDaemonMutationsChain.contains`"""
		return self.contains(query)

	def size(self, query: dict | Callable[[dict], bool] = None) -> int:
		"""Returns the amount of all documents in the collection"""
		if query is None:
			return len(self._data)
		amount = 0
		for doc in self._data:
			if PetUtils.match(doc, query):
				amount += 1
		return amount

	def length(self, query: dict | Callable[[dict], bool] = None) -> int:
		"""Returns the amount of all documents in the collection"""
		return self.size(query)

	def list(self) -> list[dict]:
		"""
		Returns all documents stored in the collection as a list

		:returns: a list with all documents.
		"""
		return self._data[:]

	@property
	def _data(self) -> list[dict]:
		return self._apply_mutations()

	def _apply_mutations(self) -> list[dict]:
		if self._mutated_data is None:
			self._mutated_data = self._api.mutate(self.name, self.__mutations)
		return self._mutated_data

	def __mutate(self, mutation: str, args: Iterable) -> PetServiceMutationsChain:
		mutations = deepcopy(self.__mutations)
		mutations.append({"type": mutation, "args": args})
		return PetServiceMutationsChain(self.name, self._api, mutations)

	def __len__(self):
		return self.size()

	def __iter__(self) -> Iterator[dict]:
		return iter(self._data)

	def __getitem__(self, item: int) -> dict:
		return self._data[item]

	def __repr__(self):
		return f"<{self.__class__.__name__} name={self.name}>"

class PetServiceCollection(PetServiceMutationsChain):

	def __init__(self, name: str, api: API):
		super().__init__(name, api)

	def insert(self, doc: dict) -> dict:
		"""
		Insert a new document into the collection.

		:param doc: the document to insert
		:returns: inserted document
		"""
		if not isinstance(doc, dict):
			raise TypeError("Document must be of type dict")
		if "_id" in doc and self.get(doc["_id"]):
			raise QueryException("Duplicate id")
		doc = deepcopy(doc)
		if "_id" not in doc:
			doc["_id"] = str(uuid.uuid4())
		self._api.insert(self.name, doc)
		return doc

	def insert_many(self, docs: Iterable[dict] & Sized) -> list[dict]:
		"""
		Insert multiple documents into the collection.

		:param docs: an Iterable of documents to insert
		:returns: a list containing the inserted documents
		"""
		new_docs = []
		for doc in docs:
			if not isinstance(doc, dict):
				raise TypeError("Document must be of type dict")
			if "_id" in doc and self.get(doc["_id"]):
				raise QueryException("Duplicate id")
			doc = deepcopy(doc)
			if "_id" not in doc:
				doc["_id"] = str(uuid.uuid4())
			new_docs.append(doc)
		self._api.insert_many(self.name, new_docs)
		return new_docs

	def update_one(self, update: dict, query: str | dict = None) -> None:
		if isinstance(query, str):
			query = {"_id": query}
		elif not isinstance(query, dict):
			raise QueryException("Invalid query type: query should be dict or str")
		self._api.update_one(self.name, update, query)

	def update(self, update: dict, query: dict | list[str | dict] = None) -> None:
		if isinstance(query, list):
			if all(isinstance(item, str) for item in query):
				query = {"_id": {"$in": query}}
			elif all(isinstance(item, dict) for item in query):
				query = {"$or": query}
			else:
				raise QueryException("Invalid query format: query list should contains only ids or only docs")
		if query is not None and not isinstance(query, dict):
			raise QueryException("Invalid query type: query should be dict or list")
		self._api.update(self.name, update, query)

	def remove(self, query: str | dict | list[str | dict] | PetServiceMutationsChain) -> list[dict]:
		if isinstance(query, str):
			query = {"_id": query}
		if isinstance(query, PetServiceMutationsChain):
			query = query.list()
		if isinstance(query, list):
			if all(isinstance(item, str) for item in query):
				query = {"_id": {"$in": query}}
			elif all(isinstance(item, dict) for item in query):
				query = {"$or": query}
			else:
				raise QueryException("Invalid delete query: it can only be a list of IDs or a list of docs")
		return self._api.remove(self.name, query)

	def clear(self):
		self._api.clear(self.name)

	def get(self, id: str) -> Optional[dict]:
		"""
		Search for the document with the given ``id``.

		:param id: document's id to search
		:returns: the document with the given ``id`` or ``None``
		"""
		return self.find({"_id": id})

	def _apply_mutations(self) -> list:
		self._mutated_data = None
		return super()._apply_mutations()
